@@ -70,6 +70,8 @@ static size_t grapheme_cache_collisions = 0;
 #endif
 #endif
 
+static int cpwidth(uint32_t cp);
+
 void fcft_log_init(enum fcft_log_colorize _colorize, bool _do_syslog,
                    enum fcft_log_class _log_level);
 
@@ -150,13 +152,20 @@ struct font_priv {
 #endif
 
     tll(struct fallback) fallbacks;
-    enum fcft_emoji_presentation emoji_presentation;
+
+    struct fcft_font_options options;
     size_t ref_counter;
 };
 
 /* Global font cache */
 struct fcft_font_cache_entry {
     uint64_t hash;
+
+    char **names;
+    char *attributes;
+    size_t name_count;
+    struct fcft_font_options options;
+
     struct font_priv *font;
 
     int waiters;
@@ -506,10 +515,27 @@ static FcPattern *
 base_pattern_from_name(const char *name, FcFontSet **set)
 {
     /* Fontconfig fails to parse floating point values unless locale
-     * is e.g C, or en_US.UTF-8 */
-    assert(strcmp(setlocale(LC_NUMERIC, NULL), "C") == 0);
+     * (LC_NUMERIC) is e.g C, or en_US.UTF-8 */
+
+    locale_t c_locale = newlocale(LC_NUMERIC_MASK, "C", (locale_t)0);
+    if (c_locale == (locale_t)0) {
+        LOG_ERRNO("%s: failed to create temporary locale", name);
+        return NULL;
+    }
+
+    locale_t old_locale = uselocale(c_locale);
+    if (old_locale == (locale_t)0) {
+        LOG_ERRNO("%s: failed to set temporary C locale", name);
+        freelocale(c_locale);
+        return NULL;
+    }
+
     FcPattern *pattern = FcNameParse((const unsigned char *)name);
 
+    /* Restore locale */
+    uselocale(old_locale);
+    freelocale(c_locale);
+
     if (pattern == NULL) {
         LOG_ERR("%s: failed to lookup font", name);
         return NULL;
@@ -579,7 +605,7 @@ instantiate_pattern(FcPattern *pattern, double req_pt_size, double req_px_size,
     }
 
     FcChar8 *full_name = NULL;
-    if (FcPatternGetString(pattern, FC_FULLNAME, face_index, &full_name) != FcResultMatch)
+    if (FcPatternGetString(pattern, FC_FULLNAME, 0, &full_name) != FcResultMatch)
         LOG_WARN("failed to get full font name");
 
     mtx_lock(&ft_lock);
@@ -592,9 +618,17 @@ instantiate_pattern(FcPattern *pattern, double req_pt_size, double req_px_size,
         return false;
     }
 
-    if ((ft_err = FT_Set_Pixel_Sizes(ft_face, 0, round(pixel_size))) != FT_Err_Ok) {
-        LOG_ERR("%s: failed to set character size: %s",
-                face_file, ft_error_string(ft_err));
+    /*
+     * Using FT_Set_Char_Size() instead of FT_Set_Pixel_Sizes() allows
+     * us to use "fractional" pixel values. This _does_ make a
+     * difference.
+     *
+     * Since FT_Set_Char_Size() expects a size in *points*, not
+     * pixels, we use a DPI of 72. This means a pixel have the same
+     * size as a point.
+     */
+    if ((ft_err = FT_Set_Char_Size(ft_face, pixel_size * 64., 0, 72, 72))) {
+        LOG_ERR("%s: failed to set character size: %s", face_file, ft_error_string(ft_err));
         goto err_done_face;
     }
 
@@ -837,8 +871,8 @@ instantiate_pattern(FcPattern *pattern, double req_pt_size, double req_px_size,
             font->metrics.height, font->metrics.ascent, font->metrics.descent,
             font->metrics.max_advance.x, features);
 #else
-    LOG_INFO("%s: size=%.2fpt/%dpx, dpi=%.2f%s",
-             font->path, size, (int)round(pixel_size), dpi, features);
+    LOG_INFO("%s: size=%.2fpt/%.2fpx, dpi=%.2f%s",
+             font->path, size, pixel_size, dpi, features);
 #endif
 
     return true;
@@ -872,7 +906,8 @@ sdbm_hash(const char *s)
 }
 
 static uint64_t
-font_hash(size_t count, const char *names[static count], const char *attributes)
+font_hash(size_t count, const char *names[static count], const char *attributes,
+          const struct fcft_font_options *options)
 {
     uint64_t hash = 0;
     for (size_t i = 0; i < count; i++)
@@ -881,12 +916,52 @@ font_hash(size_t count, const char *names[static count], const char *attributes)
     if (attributes != NULL)
         hash ^= sdbm_hash(attributes);
 
+    _Static_assert(sizeof(*options) % 8 == 0, "struct fcft_font_options not a multiple of 8 bytes");
+
+    for (size_t i = 0; i < sizeof(*options) / 8; i++) {
+        const uint64_t v = ((const uint16_t *)options)[i];
+        hash ^= v;
+    }
+
     return hash;
 }
 
+static const struct fcft_font_options default_font_options = {
+    .emoji_presentation = FCFT_EMOJI_PRESENTATION_DEFAULT,
+    .color_glyphs = {
+        .srgb_decode = false,
+        .format = PIXMAN_a8r8g8b8,
+    },
+
+    .reserved = 0,
+};
+
+FCFT_EXPORT struct fcft_font_options *
+fcft_font_options_create(void)
+{
+    struct fcft_font_options *ret = calloc(1, sizeof(*ret));
+    *ret = default_font_options;
+    return ret;
+}
+
+FCFT_EXPORT void
+fcft_font_options_destroy(struct fcft_font_options *options)
+{
+    free(options);
+}
+
 FCFT_EXPORT struct fcft_font *
 fcft_from_name(size_t count, const char *names[static count],
                const char *attributes)
+{
+    return fcft_from_name2(count, names, attributes, NULL);
+}
+
+
+FCFT_EXPORT struct fcft_font *
+fcft_from_name2(
+    size_t count, const char *names[static count], const char *attributes,
+    const struct fcft_font_options *options)
 {
     if (ft_lib == NULL) {
         LOG_ERR("fcft_init() not called");
@@ -898,7 +973,19 @@ fcft_from_name(size_t count, const char *names[static count],
     if (count == 0)
         return NULL;
 
-    uint64_t hash = font_hash(count, names, attributes);
+    if (options == NULL)
+        options = &default_font_options;
+
+#if 0
+    if (options->color_glyphs.format != PIXMAN_a8r8g8b8 &&
+        options->color_glyphs.format != PIXMAN_a2r10g10b10 &&
+        options->color_glyphs.format != PIXMAN_a2b10g10r10)
+    {
+        /* Unsupported format */
+        return NULL;
+    }
+#endif
+    uint64_t hash = font_hash(count, names, attributes, options);
     struct fcft_font_cache_entry *cache_entry = NULL;
 
     mtx_lock(&font_cache_lock);
@@ -906,6 +993,29 @@ fcft_from_name(size_t count, const char *names[static count],
         if (it->item.hash != hash)
             continue;
 
+        if (memcmp(&it->item.options, options, sizeof(*options)) != 0)
+            continue;
+
+        if (it->item.name_count != count)
+            continue;
+
+        if ((attributes == NULL && it->item.attributes != NULL) ||
+            (attributes != NULL && it->item.attributes == NULL))
+        {
+            continue;
+        }
+
+        if (attributes != NULL &&
+            strcmp(it->item.attributes, attributes) != 0)
+        {
+            continue;
+        }
+
+        for (size_t i = 0; i < count; i++) {
+            if (strcmp(it->item.names[i], names[i]) != 0)
+                continue;
+        }
+
         struct fcft_font_cache_entry *e = &it->item;
 
         if (e->font != (void *)(uintptr_t)-1) {
@@ -943,6 +1053,14 @@ fcft_from_name(size_t count, const char *names[static count],
 
     cache_entry = &tll_back(font_cache);
 
+    cache_entry->name_count = count;
+    cache_entry->names = malloc(count * sizeof(char *));
+    cache_entry->attributes = attributes != NULL ? strdup(attributes) : NULL;
+    memcpy(&cache_entry->options, options, sizeof(*options));
+
+    for (size_t i = 0; i < count; i++)
+        cache_entry->names[i] = strdup(names[i]);
+
     if (cnd_init(&cache_entry->cond) != thrd_success) {
         LOG_ERR("%s: failed to instantiate font cache condition variable", names[0]);
         tll_pop_back(font_cache);
@@ -1005,14 +1123,16 @@ fcft_from_name(size_t count, const char *names[static count],
             bool grapheme_cache_lock_failed = true;
             bool pattern_failed = true;
 
-            mtx_t lock;
-            if (mtx_init(&lock, mtx_plain) != thrd_success)
+            font = calloc(1, sizeof(*font));
+            if (font == NULL)
+                goto fc_destroy;
+
+            if (mtx_init(&font->lock, mtx_plain) != thrd_success)
                 LOG_WARN("%s: failed to instantiate mutex", name);
             else
                 lock_failed = false;
 
-            pthread_rwlock_t glyph_cache_lock;
-            if (pthread_rwlock_init(&glyph_cache_lock, NULL) != 0)
+            if (pthread_rwlock_init(&font->glyph_cache_lock, NULL) != 0)
                 LOG_WARN("%s: failed to instantiate glyph cache rwlock", name);
             else
                 glyph_cache_lock_failed = false;
@@ -1025,13 +1145,11 @@ fcft_from_name(size_t count, const char *names[static count],
             } else
                 pattern_failed = false;
 
-            font = calloc(1, sizeof(*font));
             struct glyph_priv **glyph_cache_table = calloc(
                 glyph_cache_initial_size, sizeof(glyph_cache_table[0]));
 
 #if defined(FCFT_HAVE_HARFBUZZ)
-            pthread_rwlock_t grapheme_cache_lock;
-            if (pthread_rwlock_init(&grapheme_cache_lock, NULL) != 0)
+            if (pthread_rwlock_init(&font->grapheme_cache_lock, NULL) != 0)
                 LOG_WARN("%s: failed to instantiate grapheme cache rwlock", name);
             else
                 grapheme_cache_lock_failed = false;
@@ -1046,25 +1164,27 @@ fcft_from_name(size_t count, const char *names[static count],
             /* Handle failure(s) */
             if (lock_failed || glyph_cache_lock_failed ||
                 grapheme_cache_lock_failed || pattern_failed ||
-                font == NULL || glyph_cache_table == NULL
+                glyph_cache_table == NULL
 #if defined(FCFT_HAVE_HARFBUZZ)
                 || grapheme_cache_table == NULL
 #endif
                 )
             {
                 if (!lock_failed)
-                    mtx_destroy(&lock);
+                    mtx_destroy(&font->lock);
                 if (!glyph_cache_lock_failed)
-                    pthread_rwlock_destroy(&glyph_cache_lock);
+                    pthread_rwlock_destroy(&font->glyph_cache_lock);
 #if defined(FCFT_HAVE_HARFBUZZ)
                 if (!grapheme_cache_lock_failed)
-                    pthread_rwlock_destroy(&grapheme_cache_lock);
+                    pthread_rwlock_destroy(&font->grapheme_cache_lock);
 #endif
                 if (!pattern_failed)
                     free(primary);
                 free(font);
+                font = NULL;
                 free(glyph_cache_table);
                 free(grapheme_cache_table);
+fc_destroy:
                 if (langset != NULL)
                     FcLangSetDestroy(langset);
                 FcCharSetDestroy(charset);
@@ -1075,16 +1195,13 @@ fcft_from_name(size_t count, const char *names[static count],
             }
 
             font->ref_counter = 1;
-            font->lock = lock;
-            font->glyph_cache_lock = glyph_cache_lock;
             font->glyph_cache.size = glyph_cache_initial_size;
             font->glyph_cache.count = 0;
             font->glyph_cache.table = glyph_cache_table;
-            font->emoji_presentation = FCFT_EMOJI_PRESENTATION_DEFAULT;
+            font->options = *options;
             font->public = primary->metrics;
 
 #if defined(FCFT_HAVE_HARFBUZZ)
-            font->grapheme_cache_lock = grapheme_cache_lock;
             font->grapheme_cache.size = grapheme_cache_initial_size;
             font->grapheme_cache.count = 0;
             font->grapheme_cache.table = grapheme_cache_table;
@@ -1179,7 +1296,9 @@ fcft_clone(const struct fcft_font *_font)
 
 static bool
 glyph_for_index(const struct instance *inst, uint32_t index,
-                enum fcft_subpixel subpixel, struct glyph_priv *glyph)
+                enum fcft_subpixel subpixel,
+                const struct fcft_font_options *options,
+                struct glyph_priv *glyph)
 {
     glyph->valid = false;
     glyph->subpixel = subpixel;
@@ -1309,7 +1428,9 @@ glyph_for_index(const struct instance *inst, uint32_t index,
         break;
 
     case FT_PIXEL_MODE_BGRA:
-        pix_format = PIXMAN_a8r8g8b8;
+        pix_format = options->color_glyphs.srgb_decode
+            ? PIXMAN_a8r8g8b8_sRGB
+            : PIXMAN_a8r8g8b8;
         width = bitmap->width;
         rows = bitmap->rows;
         break;
@@ -1369,7 +1490,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
         }
         break;
 
-    case FT_PIXEL_MODE_BGRA: /* PIXMAN_a8r8g8b8 */
+    case FT_PIXEL_MODE_BGRA: /* PIXMAN_a8r8g8b8 by default */
         /*
          * FreeType: blue comes *first* in memory
          * Pixman: LE: blue comes *first* in memory
@@ -1391,6 +1512,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
                 *p = (uint32_t)_a << 24 | _r << 16 | _g << 8 | _b;
             }
         }
+
         break;
 
     case FT_PIXEL_MODE_LCD: /* PIXMAN_x8r8g8b8 */
@@ -1444,6 +1566,51 @@ glyph_for_index(const struct instance *inst, uint32_t index,
              pix_format, width, rows, (uint32_t *)data, stride)) == NULL)
         goto err;
 
+    if (bitmap->pixel_mode == FT_PIXEL_MODE_BGRA &&
+        options->color_glyphs.format != pix_format)
+    {
+        /*
+         * Backing buffer is currently in sRGB, a8r8g8b8. Need to
+         * decode and/or convert to requested output format.
+         */
+        assert(pix_format == PIXMAN_a8r8g8b8 ||
+               pix_format == PIXMAN_a8r8g8b8_sRGB);
+
+        /* Old pixman image + backing buffer will be replaced */
+        pixman_image_t *a8r8g8b8 = pix;
+        uint8_t *a8_data = data;
+
+        /* New backing buffer */
+        pix_format = options->color_glyphs.format;
+        stride = stride_for_format_and_width(pix_format, width);
+        data = malloc(rows * stride);
+
+        if (data == NULL) {
+            data = a8_data;
+            pix = a8r8g8b8;
+            goto err;
+        }
+
+        /* User requested output format */
+        pix = pixman_image_create_bits_no_clear(
+            pix_format, width, rows, (uint32_t *)data, stride);
+
+        if (pix == NULL) {
+            free(data);
+            data = a8_data;
+            pix = a8r8g8b8;
+            goto err;
+        }
+
+        /* "copy" a8r8g8b8 source to user requested output
+           format. This also decodes from sRGB, if requested */
+        pixman_image_composite32(
+            PIXMAN_OP_SRC, a8r8g8b8, NULL, pix, 0, 0, 0, 0, 0, 0, width, rows);
+
+        pixman_image_unref(a8r8g8b8);
+        free(a8_data);
+    }
+
     pixman_image_set_component_alpha(
         pix,
         bitmap->pixel_mode == FT_PIXEL_MODE_LCD ||
@@ -1514,7 +1681,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
         int scaled_rows = rows / (1. / inst->pixel_size_fixup);
         int scaled_stride = stride_for_format_and_width(pix_format, scaled_width);
 
-        if (pix_format == PIXMAN_a8r8g8b8) {
+        if (bitmap->pixel_mode == FT_PIXEL_MODE_BGRA) {
             uint8_t *scaled_data = malloc(scaled_rows * scaled_stride);
             if (scaled_data == NULL)
                 goto err;
@@ -1561,6 +1728,7 @@ glyph_for_index(const struct instance *inst, uint32_t index,
             },
             .width = width,
             .height = rows,
+            .is_color_glyph = bitmap->pixel_mode == FT_PIXEL_MODE_BGRA,
         },
         .subpixel = subpixel,
         .valid = true,
@@ -1578,7 +1746,9 @@ err:
 
 static bool
 glyph_for_codepoint(const struct instance *inst, uint32_t cp,
-                    enum fcft_subpixel subpixel, struct glyph_priv *glyph)
+                    enum fcft_subpixel subpixel,
+                    const struct fcft_font_options *options,
+                    struct glyph_priv *glyph)
 {
     FT_UInt idx = -1;
 
@@ -1607,9 +1777,9 @@ glyph_for_codepoint(const struct instance *inst, uint32_t cp,
     if (idx == (FT_UInt)-1)
         idx = FT_Get_Char_Index(inst->face, cp);
 
-    bool ret = glyph_for_index(inst, idx, subpixel, glyph);
+    bool ret = glyph_for_index(inst, idx, subpixel, options, glyph);
     glyph->public.cp = cp;
-    glyph->public.cols = wcwidth(cp);
+    glyph->public.cols = cpwidth(cp);
     return ret;
 }
 
@@ -1788,8 +1958,8 @@ fcft_rasterize_char_utf32(struct fcft_font *_font, uint32_t cp,
     bool force_emoji_presentation = false;
     bool enforce_presentation_style = emoji != NULL;
 
-    if (emoji != NULL) {
-        switch (font->emoji_presentation) {
+    if (emoji != NULL && !emoji->component) {
+        switch (font->options.emoji_presentation) {
         case FCFT_EMOJI_PRESENTATION_TEXT:
             force_text_presentation = true;
             force_emoji_presentation = false;
@@ -1852,7 +2022,7 @@ search_fonts:
         }
 
         assert(it->item.font != NULL);
-        got_glyph = glyph_for_codepoint(it->item.font, cp, subpixel, glyph);
+        got_glyph = glyph_for_codepoint(it->item.font, cp, subpixel, &font->options, glyph);
         no_one = false;
         break;
     }
@@ -1870,7 +2040,7 @@ search_fonts:
         struct instance *inst = tll_front(font->fallbacks).font;
 
         assert(inst != NULL);
-        got_glyph = glyph_for_codepoint(inst, cp, subpixel, glyph);
+        got_glyph = glyph_for_codepoint(inst, cp, subpixel, &font->options, glyph);
     }
 
     assert(*entry == NULL);
@@ -2004,7 +2174,7 @@ font_for_grapheme(struct font_priv *font,
                                      cluster[i] < emoji->cp + emoji->count));
 
             if (enforce_presentation_style &&
-                emoji != NULL &&
+                emoji != NULL && !emoji->component &&
                 (i + 1 >= len || (cluster[i + 1] != 0xfe0e &&
                                   cluster[i + 1] != 0xfe0f))) {
                 /*
@@ -2015,7 +2185,7 @@ font_for_grapheme(struct font_priv *font,
                 bool force_text_presentation = false;
                 bool force_emoji_presentation = false;
 
-                switch (font->emoji_presentation) {
+                switch (font->options.emoji_presentation) {
                 case FCFT_EMOJI_PRESENTATION_TEXT:
                     force_text_presentation = true;
                     force_emoji_presentation = false;
@@ -2192,7 +2362,7 @@ fcft_rasterize_grapheme_utf32(struct fcft_font *_font,
     for (size_t i = 0; i < len; i++) {
         if (cluster[i] == 0xfe0f)
             min_grapheme_width = 2;
-        grapheme_width += wcwidth(cluster[i]);
+        grapheme_width += cpwidth(cluster[i]);
     }
 
     LOG_DBG("length: %u", hb_buffer_get_length(inst->hb_buf));
@@ -2215,7 +2385,7 @@ fcft_rasterize_grapheme_utf32(struct fcft_font *_font,
 
         struct glyph_priv *glyph = malloc(sizeof(*glyph));
         if (glyph == NULL ||
-            !glyph_for_index(inst, info[i].codepoint, subpixel, glyph))
+            !glyph_for_index(inst, info[i].codepoint, subpixel, &font->options, glyph))
         {
             assert(glyph == NULL || !glyph->valid);
             free(glyph);
@@ -2226,7 +2396,7 @@ fcft_rasterize_grapheme_utf32(struct fcft_font *_font,
 
         assert(info[i].cluster < len);
         glyph->public.cp = cluster[info[i].cluster];
-        glyph->public.cols = wcwidth(glyph->public.cp);
+        glyph->public.cols = cpwidth(glyph->public.cp);
 
 #if 0
         LOG_DBG("grapheme: x: advance: %d -> %d, offset: %d -> %d",
@@ -2308,7 +2478,8 @@ static bool
 rasterize_partial_run(struct text_run *run, const struct instance *inst,
                       const uint32_t *text, size_t len,
                       size_t run_start, size_t run_len,
-                      enum fcft_subpixel subpixel)
+                      enum fcft_subpixel subpixel,
+                      const struct fcft_font_options *options)
 {
     hb_buffer_add_utf32(inst->hb_buf, text, len, run_start, run_len);
     hb_buffer_guess_segment_properties(inst->hb_buf);
@@ -2342,14 +2513,14 @@ rasterize_partial_run(struct text_run *run, const struct instance *inst,
         if (glyph == NULL)
             return false;
 
-        if (!glyph_for_index(inst, info->codepoint, subpixel, glyph)) {
+        if (!glyph_for_index(inst, info->codepoint, subpixel, options, glyph)) {
             free(glyph);
             continue;
         }
 
         assert(info->cluster < len);
         glyph->public.cp = text[info->cluster];
-        glyph->public.cols = wcwidth(glyph->public.cp);
+        glyph->public.cols = cpwidth(glyph->public.cp);
 
         /* TODO: can’t reference font data, since the font may be
          * free:d before the text-run (and thus all the text-run’s
@@ -2526,7 +2697,7 @@ fcft_rasterize_text_run_utf32(
 
         bool ret = rasterize_partial_run(
             &run, prun->inst, (const uint32_t *)text, len,
-            prun->start, prun->len, subpixel);
+            prun->start, prun->len, subpixel, &font->options);
 
         hb_buffer_clear_contents(prun->inst->hb_buf);
         if (!ret)
@@ -2628,6 +2799,12 @@ fcft_destroy(struct fcft_font *_font)
             mtx_unlock(&font->lock);
 
             cnd_destroy(&it->item.cond);
+
+            for (size_t i = 0; i < it->item.name_count; i++)
+                free(it->item.names[i]);
+            free(it->item.names);
+            free(it->item.attributes);
+
             tll_remove(font_cache, it);
             break;
         }
@@ -2821,5 +2998,15 @@ fcft_set_emoji_presentation(struct fcft_font *_font,
                             enum fcft_emoji_presentation presentation)
 {
     struct font_priv *font = (struct font_priv *)_font;
-    font->emoji_presentation = presentation;
+    font->options.emoji_presentation = presentation;
+}
+
+static int
+cpwidth(uint32_t cp)
+{
+#if defined(FCFT_HAVE_UTF8PROC)
+    return utf8proc_charwidth((utf8proc_int32_t)cp);
+#else
+    return wcwidth((wchar_t)cp);
+#endif
 }
